/*************************************************************************
 *
 * file: vobwalker.c
 *
 * This program walks through a .VOB file, splitting it into smaller 
 * files. Each file begins with a VOBU that clains to be the first
 * in a chain (i.e., it says there are no preceding VOBUs containing
 * video), through the chain of VOBUs found by following the Next
 * VOBU with Video links, and ends when a VOBU is encountered that
 * claims there are no succeeding VOBUs.
 *
 * In practical terms, you get a .VOB file for each chapter in the
 * source material.
 *
 * This program requires random-access to the source .VOB file and
 * creates multiple .VOB output files, so it CANNOT be used as a
 * filter.
 *
 * It also assumes the source is a single .VOB file; if you have
 * multiple related files, you'll need to cat them together before
 * spitting it through vobwalker. This means you'll need at least
 * 3 times as much disk space as the source (one for the original 
 * 1GB .VOBs, one for the catted huge .VOB, and one more for the
 * split chapter .VOBs).
 *
 * Oh yeah. I use NetBSD, so this program has only been tried on
 * that. It assumes it has a uint64_t, which is the appropriate
 * size to be used as an off_t.
 *
 * This code is based on information found at 
 * http://dvd.sourceforge.net/dvdinfo/
 *
 * This program is in the public domain.
 *
 * revisions:
 *
 *    2004-06-01 anachronda@hotmail.com [rli]: Clean up and commenting of 
 *      original.
 *
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

/***
 *
 * Spin
 *
 *    Advances a spinner on stdout. The current spinner character is displayed
 *    on stdout followed by a backspace; this makes an animated spinner
 *    appear to display the program's progress.
 *
 * revisions:
 *
 *    2004-06-01 rli: original version.
 *
 * formal parameters:
 *
 *    none.
 *
 * informal parameters:
 *
 *    Initialized static variables are used to keep track of the
 *    spinner state.
 *
 * return value:
 *
 *    none.
 *
 * side effects:
 *
 *    The static spinner count is cycled; the effect of this is that for
 *    each call, the spinner is advanced a single tick.
 *
 ***/

void Spin( void )
{
  static char Spinner[ 4 ] = "|/-\\";
  static int SpinCount = 0;

  printf( "%c\010", Spinner[ SpinCount ] );
  fflush( stdout );
  SpinCount = ( SpinCount + 1 ) & 3;
}

/* ================= Pack ================
 *
 * These routines deal with pulling a PACK from the .VOB file and
 * examining it.
 *
 */

/***
 *
 * Pack_Suck
 *
 * Reads a packet from the specified file into the specified buffer. 
 *
 * revisions:
 *
 *     2004-06-01 rli: Reworked comments.
 *
 * formal parameters:
 *
 *     - From: The file from which the packet should be read.
 *
 *     - Buffer: The buffer into which the packet should be placed.
 *
 * informal parameters:
 *
 *     A packet is assumed to contain 2048 bytes.
 *
 * return value:
 *
 *     - 0: Failed. Fewer than 2048 characters were read from the file.
 *       The buffer should not be used; it does not contain a packet.
 *
 *     - 1: Success. The buffer contains a packet that may be examined.
 *
 * side effects:
 *
 *     none.
 *
 ***/

int Pack_Suck( FILE *From, unsigned char *Buffer )
{
  int BytesRead;

  BytesRead = fread( Buffer, 1, 2048, From );
  if( BytesRead < 2048 ) return 0;
  return 1;
}

/***
 *
 * Pack_IsEndCode
 *
 *     Examines the specified packet to see if is an MPEG-2 end program
 *     code. An end program code begins with the sequence 00 00 01 B9 (hex).
 *
 * revisions:
 *
 *     2004-06-01 rli: Reworked comments.
 *
 * formal parameters:
 *
 *     - Buffer: The packet to examine.
 *
 * informal parameters:
 *
 *     none.
 *
 * return value:
 *
 *     - 0: The packet is NOT an MPEG-2 end program code.
 *
 *     - 1: The packet is an MPEG-2 end program code.
 *
 * side effects:
 *
 *     none.
 *
 ***/

int Pack_IsEndCode( unsigned char *Buffer )
{
  if( ( Buffer[ 0 ] == 0 ) && ( Buffer[ 1 ] == 0 ) && 
      ( Buffer[ 2 ] == 1 ) && ( Buffer[ 3 ] == 0xb9 ) ) return 1;
  return 0;
}

/***
 *
 * Pack_IsVobu
 *
 *     Examines the specified packet to see if it's a VOBU. For our
 *     purposes, A VOBU begins with 00 00 01 BA (hex) and contains 
 *     00 00 01 BF (hex) in bytes 0x400 through 0x403.
 *
 * revisions:
 *
 *     2004-06-01 rli: Reworked comments. Added check for 000001BA at
 *       start of packet so we can't be fooled by malicious end packets.
 *
 * formal parameters:
 *
 *     - Buffer: The packet to be examined.
 *
 * informal parameters:
 *
 *     none.
 *
 * return value:
 *
 *     - 0: The packet is NOT a VOBU.
 *
 *     - 1: The packet is a VOBU.
 *
 * side effects:
 *
 *     none.
 *
 ****/

int Pack_IsVobu( unsigned char *Buffer )
{
  if( ( Buffer[ 0 ] == 0x00 ) && ( Buffer[ 1 ] == 0x00 ) &&
      ( Buffer[ 2 ] == 0x01 ) && ( Buffer[ 3 ] == 0xba ) &&
      ( Buffer[ 0x400 ] == 0x00 ) && ( Buffer[ 0x401 ] == 0x00 ) &&
      ( Buffer[ 0x402 ] == 0x01 ) && ( Buffer[ 0x403 ] == 0xbf ) ) return 1;
  return 0;
}

/* ================ Vobu ===================
 *
 * These routines deal with extracting information from a VOBU.
 *
 */

/***
 *
 * Vobu_ExtractEndAddress
 *
 *     Extracts the end address from a VOBU. The end address is
 *     specified as a relative offset block offset from the current
 *     block. Thus, a VOBU that consists of a single block would have
 *     an end address of 0 while one that consists of two blocks would
 *     have an end address of 1.
 *
 *     The end address is contained in bytes 406 through 412 of the VOBU.
 *
 * revisions:
 *
 *     2004-06-01 rli: reworked comments.
 *
 * formal parameters:
 *
 *     - Buffer: The packet from which the end address is to be
 *       extracted.
 *
 * informal parameters:
 *
 *     none.
 *
 * return value:
 *
 *     The end address extracted from the VOBU.
 *
 * side effects:
 *
 *     none.
 *
 ***/

unsigned int Vobu_ExtractEndAddress( unsigned char *Buffer )
{
  unsigned int Temp;

  Temp = ( Buffer[ 0x40f ] << 24 ) | ( Buffer[ 0x410 ] << 16 ) |
         ( Buffer[ 0x411 ] << 8 ) | Buffer[ 0x412 ];
  return Temp;
}

/***
 *
 * Vobu_ExtractNextAddress
 *
 *     Extracts the address of the next VOBU with video from the
 *     specified VOBU. The address is a relative packet count from the
 *     current packet address, except that bit 31 is set to indicate the
 *     pointer is valid. The value 0xBFFFFFFF is reserved to indicate there 
 *     is no subsequent VOBU containing video.
 *
 *     This program extracts chapters by walking a chain of VOBUs from
 *     the first VOBU in the chain until it encounters one that has no
 *     following VOBU.
 *
 * revisions:
 *
 *     2004-06-01 rli: Reworked comments.
 *
 * formal parameters:
 *
 *     - Buffer: The packet from which the next address is to be
 *       extracted.
 *
 * informal parameters:
 *
 *     none.
 *
 * return value:
 *
 *     The next address extracted from the VOBU.
 *
 *     NOTE:   The caller must remember to clear bit 31 before
 *             using the pointer!!
 *
 * side effects:
 *
 *     none.
 *
 ***/

unsigned int Vobu_ExtractNextAddress( unsigned char *Buffer )
{
  unsigned int Temp;

  Temp = ( Buffer[ 0x4f1 ] << 24 ) | ( Buffer[ 0x4f2 ] << 16 ) |
         ( Buffer[ 0x4f3 ] << 8 ) | Buffer[ 0x4f4 ];
  return Temp;
}

/***
 *
 * Vobu_IsLast
 *
 *     Examines the next address pointer from the specified VOBU to
 *     determine whether the current VOBU has a successor. If it does,
 *     it is not the last VOBU in the chapter.
 *
 *     The last packet in a chain has a next address of 0xBFFFFFFF.
 *
 * revisions:
 *
 *     2004-06-01 rli: Reworked comments.
 *
 * formal parameters:
 *
 *     - Buffer: The packet to be examined.
 *
 * informal parameters:
 *
 *     none.
 *
 * return value:
 *
 *     - 0: Either this is the last VOBU in a chain or the packet is
 *       not a VOBU.
 *
 *     - 1: The packet is a VOBU with no successor; it is the last
 *       VOBU in a chain.
 *
 * side effects:
 *
 *     none.
 *
 ***/

int Vobu_IsLast( unsigned char *Buffer )
{
  if( Pack_IsVobu( Buffer ) &&
    ( Buffer[ 0x4f1 ] == 0xbf ) && ( Buffer[ 0x4f2 ] == 0xff ) &&
    ( Buffer[ 0x4f3 ] == 0xff ) && ( Buffer[ 0x4f4 ] == 0xff ) ) return 1;
  return 0;
}

/***
 *
 * Vobu_IsFirst
 *
 *     Examines the packet to determine whether it is the first VOBU in
 *     a chapter. Such a VOBU has no preceding VOBUs with video; the
 *     pointer to the preceding VOBU contains the reserved value
 *     0xBFFFFFFF.
 *
 *     This program extracts chapters by walking a chain of VOBUs from
 *     the first VOBU in the chain until it encounters one that has no
 *     following VOBU.
 *
 * revisions:
 *
 *     2004-06-01 rli: reworked comments.
 *
 * formal parameters:
 *
 *     - Buffer: The packet to be examined.
 *
 * informal parameters:
 *
 *     none.
 *
 * return value:
 *
 *     - 0: The packet is not the first VOBU in a chain; either it is
 *       not a VOBU or the predecessor address is not the reserved
 *       value that indicates there is no predecessor.
 *
 *     - 1: The packet is the first VOBU in a chain.
 *
 * side effects:
 *
 *     none.
 *
 ***/

int Vobu_IsFirst( unsigned char *Buffer )
{
  if( Pack_IsVobu( Buffer ) &&
    ( Buffer[ 0x595 ] == 0xbf ) && ( Buffer[ 0x596 ] == 0xff ) &&
    ( Buffer[ 0x597 ] == 0xff ) && ( Buffer[ 0x598 ] == 0xff ) ) return 1;
  return 0;
}

/* ============== Blow =================
 *
 * These routines deal with copying packets to an output file.
 *
 */

/* Blow_CurrentFile
 *
 *     The file currently open for output. Packets are written to this
 *     file.
 *
 *     If zero, there is no file currently open.
 */

FILE *Blow_CurrentFile = NULL;

/* Blow_FileNumber
 *
 *     The number of the current chapter. This is combined with
 *     Blow_Format to form the name of the chapter file.
 */

int Blow_FileNumber = 0;

/* Blow_FileName
 *
 *     The name of the current chapter file.
 */

char Blow_FileName[ 80 ];

/* Blow_Format
 *
 *     The format string through which Blow_FileNumber is printed to
 *     form Blow_FileName.
 */

char *Blow_Format;

/* Blow_DefaultFormat
 *
 *     The format used to construct a chapter file name from the file
 *     number IF a different format is NOT specified on the command
 *     line.
 */

char Blow_DefaultFormat[] = "chap%d.vob";

/***
 *
 * Blow_EndFile
 *
 *     Closes the current chapter file, if one is open.
 *
 * revisions:
 *
 *     2004-06-01 rli: Reworked comment; check for close failure. Exit
 *       with EXIT_FAILURE on close failure.
 *
 * formal parameters:
 *
 *     none.
 *
 * informal parameters:
 *
 *     - Blow_CurrentFile: The file that is currently open. If zero,
 *       there is no such file.
 *
 *     - Blow_FileName: The name of the current file. This is used to
 *       display an error message if the close fails.
 *
 * return value:
 *
 *     none.
 *
 * side effects:
 *
 *     If the close fails, the program is terminated.
 *
 ***/

void Blow_EndFile( void )
{
  if( Blow_CurrentFile != NULL ) {
    if( fclose( Blow_CurrentFile ) == EOF ) {
      fprintf( stderr, "Blow_EndFile: Could not close %s\n",
        Blow_FileName );
      perror( NULL );
      exit( EXIT_FAILURE );
    }
    Blow_CurrentFile = NULL;
  }
}

/***
 *
 * Blow_NextFile
 *
 *     Closes the current output file, if one is open, and opens the
 *     next one.
 *
 *     The file name is displayed on stdout after it is opened. It is
 *     followed by a space to provide a good place to display a spinner.
 *      
 * revisions:
 *
 *     2004-06-01 rli: Reworked comments. Use Blow_Format and
 *       snprintf to construct the filename. Display errno using
 *       perror and exit with EXIT_FAILURE if open fails.
 *
 * formal parameters:
 *
 *     none.
 *
 * informal parameters:
 *
 *     - Blow_CurrentFile: The current output file.
 *
 *     - Blow_FileNumber: The current file number. This is incremented.
 *
 *     - Blow_FileName: The current output file name. This is generated
 *       by printing Blow_FileNumber through Blow_Format.
 *
 *     - Blow_Format: A format string that may be used, with
 *       Blow_FileNumber, to construct a file name.
 *
 * return value:
 *
 *     none.
 *
 * side effects:
 *
 *     If the next file could not be opened, the program is aborted.
 *
 ***/

void Blow_NextFile( void )
{
  Blow_EndFile();
  Blow_FileNumber++;
  snprintf( Blow_FileName, sizeof( Blow_FileName ), Blow_Format, 
    Blow_FileNumber );
  Blow_CurrentFile = fopen( Blow_FileName, "wb" );
  if( Blow_CurrentFile == 0 ) {
    fprintf( stderr, "Blow_NextFile: Couldn't create %s\n", Blow_FileName );
    perror( NULL );
    exit( EXIT_FAILURE );
  }
  printf( "%s ", Blow_FileName );
  fflush( stdout );
}

/***
 *
 * Blow_WritePacket
 *
 *     Writes a packet to the current output file. If a file is not
 *     open, one is created.
 *
 * revisions:
 *
 *     2004-06-01 rli: Reworked comments, renamed. Display stderr
 *       using perror and exit with EXIT_FAILURE if write fails.
 *
 * formal parameters:
 *
 *     Buffer: The packet to be written.
 *
 * informal parameters:
 *
 *     - Blow_CurrentFile: The current output file. If zero, there is
 *       no output file open.
 *
 * return value:
 *
 *     none. 
 *
 * side effects:
 *
 *     If a file is created, several other Blow_ globals are updated
 *     by Blow_NextFile.
 *
 *     If the write fails, the program is aborted.
 *
 ***/

void Blow_WritePacket( unsigned char *Buffer )
{
  int BytesWritten;

  if( Blow_CurrentFile == 0 ) Blow_NextFile();
  BytesWritten = fwrite( Buffer, 1, 2048, Blow_CurrentFile );
  if( BytesWritten != 2048 ) {
    fprintf( stderr, "Blow_WritePacket: Short write! Wrote %d expected 2048\n",
      BytesWritten );
    perror( NULL );
    exit( EXIT_FAILURE );
  }
}

/* ================= Chapter =====================
 *
 * These routines deal with following a chain of VOBUs.
 */

/***
 *
 * Chapter_BlowVobu
 *
 *     Blows the data associated with a specific VOBU into an
 *     output file.
 *
 *
 * revisions:
 *
 *     2004-06-01 rli: Reworked comments. Rework failure paths.
 *
 * formal parameters:
 *
 *     - VobFile: The file to which the data should be written.
 *
 *     - Address: The address of the VOBU to be copied to the file.
 *       Since a VOBU contains 2048 bytes, this is multiplied by
 *       2048 to form the file address of the first block in the
 *       VOBU.
 *
 * informal parameters:
 *
 *     none.
 *
 * return value:
 *
 *     none.
 *
 * side effects:
 *
 *     The program is aborted if an error occurs.
 *
 ***/

void Chapter_BlowVob( FILE *VobFile, uint64_t Address )
{
  unsigned char PacketBuffer[ 2048 ];
  unsigned int BlockCount;

  /* Seek to the first block of the VOBU. */

  if( fseeko( VobFile, Address * 2048, SEEK_SET == -1 ) ) {
    fprintf( stderr, "Chapter_BlowVob: Cout not seek to %qu\n", Address );
    perror( NULL );
    exit( EXIT_FAILURE );
  }

  /* Fetch the VOBU so we can extract the block count. */

  if( Pack_Suck( VobFile, PacketBuffer ) == 0 ) {
    fprintf( stderr, "Chapter_BlowVob: Couldn't suck packet\n" );
    perror( NULL );
    exit( EXIT_FAILURE );
  }

  /* Sanity check: Verify this packet is a VOBU. */

  if( Pack_IsVobu( PacketBuffer ) == 0 ) {
    fprintf( stderr, 
      "Chapter_BlowVob: Seeked to something that isn't a VOBU\n" );
    exit( EXIT_FAILURE );
  }

  /* Extract the relative address of the end packet from the VOBU. Since
   * we want a counter that we can decrement after we write each packet,
   * we'll need to add one to it.
   */

  BlockCount = Vobu_ExtractEndAddress( PacketBuffer ) + 1;

  /* Copy packets to the output file until we either encounter an error
   * or have copied them all.
   */

  do {
    Blow_WritePacket( PacketBuffer );
 
    if( Pack_Suck( VobFile, PacketBuffer ) == 0 ) {
      fprintf( stderr, "Chapter_BlowVob: Couldn't suck packet\n" );
      perror( NULL );
      exit( EXIT_FAILURE );
    }

    BlockCount--;
  } while( BlockCount > 0 );
}

/***
 *
 * Chapter_Walk
 *
 *     Walks through the VOBUs in a chapter, copying their associated
 *     data to a file, by following the next vob link.
 *
 *     The spinner is advanced for each VOBU.
 *
 * revisions:
 *
 *     2004-06-01 rli: Reworked comments, trimmed display to the
 *       spinner, miscellaneous restructuring.
 *
 * formal parameters:
 *
 *     - VobFile: The file to which data should be copied.
 *
 *     - Address: The address of the first VOBU in the chapter. This is
 *       a block address, so it needs to be multiplied by 2048 to form
 *       a byte address in the file.
 *
 * informal parameters:
 *
 *    none.
 *
 * return value:
 *
 *     none.
 *
 * side effects:
 *
 *     The next output file is opened.
 *
 *     If an error occurs, the program is aborted.
 *
 ***/

void Chapter_Walk( FILE *VobFile, uint64_t Address )
{
  unsigned char PacketBuffer[ 2048 ];

  /* Create the next output file. */

  Blow_NextFile();

  do {

    /* Seek to the next VOBU in the chapter. */

    if( fseeko( VobFile, Address * 2048, SEEK_SET ) == -1 ) {
      fprintf( stderr, "Chapter_Walk: could not seek to %qu\n", Address );
      perror( NULL );
      exit( EXIT_FAILURE );
    }

    /* Read the VOBU */

    if( Pack_Suck( VobFile, PacketBuffer ) == 0 ) {
      fprintf( stderr, "Chapter_Walk: Couldn't suck packet\n" );
      perror( NULL );
      exit( EXIT_FAILURE );
    }

    /* Sanity check that we've found a VOBU. */

    if( !Pack_IsVobu( PacketBuffer ) ) {
      fprintf( stderr, 
        "Chapter_Walk: Seeked to something that's not a VOBU\n" );
      perror( NULL );
      exit( EXIT_FAILURE );
    }

    /* We have our VOBU. Advance the spinner and copy its associated
     * data into the output file.
     */

    Spin();
    Chapter_BlowVob( VobFile, Address );

    /* Move on to the next VOBU in the chaper. The next address is a
     * relative packet count with bit 31 set, so we need to clear bit
     * 31 before we add it to the current packet address.
     */
 
    Address += ( Vobu_ExtractNextAddress( PacketBuffer ) & 0x7fffffff );

  } while( !Vobu_IsLast( PacketBuffer ) );

  /* We're done with this chapter. Close the file and erase the spinner.  */

  Blow_EndFile();
  printf( " \n" );
}

/***
 *
 * main
 *
 *     This program walks through a .VOB file, splitting it into smaller 
 *     files. Each file begins with a VOBU that clains to be the first
 *     in a chain (i.e., it says there are no preceding VOBUs containing
 *     video), through the chain of VOBUs found by following the Next
 *     VOBU with Video links, and ends when a VOBU is encountered that
 *     claims there are no succeeding VOBUs.
 *
 * revisions:
 *
 *     2004-06-01 rli: General rework.
 *
 * formal parameters:
 *
 *     - ArgC: The argument count. This should be either two or three. If
 *       two, only a vob file name has been specified. If three, both a
 *       vob file name and a format string for generating chapter file
 *       names have been specified.
 *
 *     - ArgV: Command-line arguments. This program expects:
 *
 *       - ArgV[ 1 ]: The name of the .VOB file to be split. Since this
 *         program requires random access to the .VOB file, it CANNOT
 *         read the file from stdin.
 *
 *       - ArgV[ 2 ]: A format string to be used to generate chapter
 *         filenames. A chapter filename is generated by printing
 *         the filenumber through this string.
 *
 *         If this parameter is not specified, Blow_DefaultFormat is used.
 *
 * informal parameters:
 *
 *     - Blow_DefaultFormat: The format string that should be used to
 *       generate filenames if another format is not specified.
 *
 * return value:
 *
 *     - EXIT_SUCCESS. Errors cause abortion via exit() rather than a
 *       return.
 *
 * side effects:
 *
 *     Since chapter files are created, the Blow_ globals are updated.
 *
 *     If an error occurs, the program is aborted.
 *
 ***/

int main( int ArgC, char **ArgV )
{
  unsigned char PacketBuffer[ 2048 ];
  uint64_t Address = 0;
  uint64_t NextAddress;
  FILE *VobFile;

  /* We can deal with either two or three command line arguments. */

  if( ( ArgC < 2 ) || ( ArgC > 3 ) ) {
    fprintf( stderr, "usage: vob-chase vobfile [printf-format-string]\n" );
    exit( EXIT_FAILURE );
  }

  /* Set the format string used to generate a file name. This is 
   * Blow_DefaultFormat unless a different one has been specified on
   * the command line.
   */

  if( ArgC == 3 ) {
    Blow_Format = ArgV[ 2 ];
  } else {
    Blow_Format = Blow_DefaultFormat;
  }

  /* Open the .VOB file we're supposed to split. */

  VobFile = fopen( ArgV[ 1 ], "rb" );
  if( VobFile == 0 ) {
    fprintf( stderr, "main: couldn't open %s\n", ArgV[ 1 ] );
    perror( NULL );
    exit( EXIT_FAILURE );
  }

  while(1) {

    /* Seek to the next VOBU. */

    if( fseeko( VobFile, Address * 2048, SEEK_SET ) == -1 ) {
      fprintf( stderr, "main: could not seek to %qu\n", Address );
      perror( NULL );
      exit( EXIT_FAILURE );
    }

    /* Read the VOBU to which we've just seeked. Inability to read
     * the VOBU is an error; we should see an end code at the end
     * of the file.
     */

    if( Pack_Suck( VobFile, PacketBuffer ) == 0 ) {
      fprintf( stderr, "main: Couldn't suck packet\n" );
      perror( NULL );
      exit( EXIT_FAILURE );
    }

    /* If we've made it to the end code, we've processed the entire
     * file; we're done, and successfully to boot.
     */

    if( Pack_IsEndCode( PacketBuffer ) ) {
      Blow_EndFile();
      printf( "End code found.\n" );
      return EXIT_SUCCESS;
    }

    /* We have a packet that is not an end code. Double-check that it's
     * a VOBU.
     *
     * I'm not entirely certain that this should be an error since we're
     * examining every packet in the file, whether it's part of a VOBU
     * chain or not. It may be the case that other packets are included
     * that are never referenced by a VOBU chain and therefore never
     * seen by a DVD player.
     *
     * Until I'm convinced that this may actually happen I'll treat it
     * as an error. If you're having trouble, feel free to omit this
     * test.
     */

    if( !Pack_IsVobu( PacketBuffer ) ) {
      fprintf( stderr, "main: Seeked to something that's not a VOBU\n" );
      exit( EXIT_FAILURE );
    }

    /* We have a VOBU. If it's the first VOBU of a chain, we need to
     * extract the chapter. 
     */

    if( Vobu_IsFirst( PacketBuffer ) ) {
      Chapter_Walk( VobFile, Address );
    }

    /* We're done with this VOBU. Advance the spinner and move on to
     * the next one (via the end address pointer rather than the
     * next with video pointer; this will move us sequentially
     * the file).
     */

    Spin();
    Address += ( Vobu_ExtractEndAddress( PacketBuffer ) + 1 );
  }

}
